package com.cedarsoft.spring.rcp.beans.factory;

import com.cedarsoft.spring.SpringSupport;
import com.cedarsoft.spring.rcp.beans.BeanAgent;
import com.cedarsoft.spring.rcp.beans.EditorBasedBeanAgent;
import com.cedarsoft.spring.rcp.editor.FormBasedEditor;
import com.cedarsoft.history.DelegatingObjectAccess;
import com.cedarsoft.history.ElementsChangedEvent;
import com.cedarsoft.history.ElementsListener;
import com.cedarsoft.history.ObservableObjectAccess;
import org.jetbrains.annotations.NotNull;
import org.springframework.richclient.application.docking.jide.editor.EditorDescriptor;
import org.springframework.richclient.application.docking.jide.editor.CloseEditorEvent;

import java.lang.ref.WeakReference;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

/**
 * UI Factory for business objects.
 * This factory is configured using a {@link UiAccessBuilderConfiguration}.
 * <p/>
 * This factory works based on extended tables, editors and views.
 * It should be the right thing for most of the default beans
 */
public class EditorBasedUiAccessBuilder<T> extends UiAccessBuilder<T> {
  public EditorBasedUiAccessBuilder( @NotNull UiAccessBuilderConfiguration<T> configuration ) {
    super( configuration );
  }

  @Override
  @NotNull
  protected BeanAgent<T> createAgent() {
    return new EditorBasedBeanAgent<T>( getConfiguration().getBeanFactory() );
  }

  @NotNull
  @Override
  public UiAccessBuilderConfiguration<T> getConfiguration() {
    return ( UiAccessBuilderConfiguration<T> ) super.getConfiguration();
  }

  @NotNull
  public EditorDescriptor getEditorDescriptor(@NotNull final ObservableObjectAccess<T> objectAccess) {
    EditorDescriptor editorDescriptor = new EditorDescriptor();

    editorDescriptor.setEditorClass( FormBasedEditor.class );
    //    editorDescriptor.setBeanName( getConfiguration().getIdStrategy().getBaseId() );

    Map<Object, Object> map = new HashMap<Object, Object>();
    map.put( FormBasedEditor.PROPERTY_ID_STRATEGY, getConfiguration().getIdStrategy() );
    map.put( FormBasedEditor.PROPERTY_FORM_FACTORY, getConfiguration().getFormFactory() );
    map.put( FormBasedEditor.PROPERTY_OBJECT_COMMIT, objectAccess );
    map.put( FormBasedEditor.PROPERTY_LABEL_PROVIDER, getLabelProvider() );

    FormBasedEditor.InitializerCallback<T> initializerCallback = new FormBasedEditor.InitializerCallback<T>() {
      @Override
      public void isInitializing( @NotNull final FormBasedEditor<T> editor ) {
        if ( objectAccess instanceof DelegatingObjectAccess ) {
          ObservableObjectAccess<T> currentDelegate = ( ( DelegatingObjectAccess<T> ) objectAccess ).getCurrentDelegate();
          if ( currentDelegate == null ) {
            throw new IllegalStateException( "No delegate available - impossible" );
          }
          currentDelegate.addElementListener( new EditorCloseListener<T>( editor ) );
        } else {
          objectAccess.addElementListener( new EditorCloseListener<T>( editor ) );
        }
      }
    };

    map.put( FormBasedEditor.PROPERTY_INITIALIZER_CALLBACK, initializerCallback );
    editorDescriptor.setEditorProperties( map );

    return editorDescriptor;
  }

  private class EditorCloseListener<T> implements ElementsListener<T> {
    @NotNull
    private final WeakReference<FormBasedEditor<T>> reference;

    private EditorCloseListener( @NotNull FormBasedEditor<T> reference ) {
      this.reference = new WeakReference<FormBasedEditor<T>>( reference );
    }

    @Override
    public void elementsDeleted( @NotNull ElementsChangedEvent<? extends T> event ) {
      FormBasedEditor<T> editor = reference.get();
      if ( editor == null ) {
        event.getSource().removeElementListener( this );
        return;
      }

      T object = editor.getEditorObject();
      List<? extends T> elements = event.getElements();
      if ( elements.contains( object ) ) {
        //Revert, so that the editor can be closed...
        editor.getForm().revert();
        SpringSupport.INSTANCE.publishEvent( new CloseEditorEvent( editor.getId() ) );
      }
    }

    @Override
    public void elementsAdded( @NotNull ElementsChangedEvent<? extends T> event ) {
    }

    @Override
    public void elementsChanged( @NotNull ElementsChangedEvent<? extends T> event ) {
    }
  }
}