package com.cedarsoft.spring.rcp.beans.factory;

import com.cedarsoft.history.ObservableObjectAccess;
import com.cedarsoft.spring.rcp.beans.AccessAgent;
import com.cedarsoft.spring.rcp.beans.BeanAgent;
import com.cedarsoft.spring.rcp.beans.BeanUiAccess;
import com.cedarsoft.spring.rcp.beans.BeansDisplay;
import com.cedarsoft.spring.rcp.beans.DefaultBeanUiAccess;
import com.cedarsoft.spring.rcp.beans.ObjectTableBasedAccessAgent;
import com.cedarsoft.spring.rcp.beans.view.ViewBasedBeansDisplay;
import com.cedarsoft.spring.rcp.forms.FormFactory;
import com.cedarsoft.spring.rcp.id.IdStrategy;
import com.cedarsoft.spring.rcp.tbpanel.DefaultObjectTablePanelFactory;
import com.cedarsoft.spring.rcp.tbpanel.ObjectTablePanelFactory;
import com.cedarsoft.spring.rcp.view.ObjectTableView;
import com.cedarsoft.spring.rcp.view.ViewConfiguration;
import com.cedarsoft.spring.rcp.views.ExtendedView;
import org.jetbrains.annotations.NotNull;
import org.springframework.richclient.selection.binding.support.LabelProvider;

import java.util.HashMap;
import java.util.Map;

/**
 * Abstract base class for ui access builders.
 *
 * @param <T> the type
 */
public abstract class UiAccessBuilder<T> {
  @NotNull
  protected final AbstractConfiguration<T> configuration;

  protected UiAccessBuilder( @NotNull AbstractConfiguration<T> configuration ) {
    this.configuration = configuration;
  }

  @NotNull
  public ViewConfiguration getViewConfiguration( @NotNull final ObservableObjectAccess<T> objectAccess ) {
    return new ViewConfiguration() {
      @Override
      @NotNull
      public Class<? extends ExtendedView> getViewClass() {
        return ObjectTableView.class;
      }

      @Override
      @NotNull
      public Map<String, Object> getViewProperties() {
        Map<String, Object> map = new HashMap<String, Object>();
        map.put( ObjectTableView.PROPERTY_OBJECT_ACCESS, objectAccess );
        map.put( ObjectTableView.PROPERTY_TABLE_BASED_PANEL_FACTORY, getObjectTableBasedFactory() );
        return map;
      }
    };
  }

  private BeansDisplay beansDisplay;

  @NotNull
  protected BeansDisplay getBeansDisplay() {
    if ( beansDisplay == null ) {
      beansDisplay = createDefaultBeansDisplay();
    }
    return beansDisplay;
  }

  @NotNull
  protected BeansDisplay createDefaultBeansDisplay() {
    return new ViewBasedBeansDisplay( getIdStrategy().createViewId() );
  }


  private AccessAgent<T> accessAgent;

  @NotNull
  protected AccessAgent<T> getAccessAgent() {
    if ( accessAgent == null ) {
      accessAgent = createDefaultAccessAgent();
    }
    return accessAgent;
  }

  @NotNull
  protected AccessAgent<T> createDefaultAccessAgent() {
    return new ObjectTableBasedAccessAgent<T>( getLabelProvider(), getObjectTableBasedFactory() );
  }

  @NotNull
  protected ObjectTablePanelFactory<T> getObjectTableBasedFactory() {
    return new DefaultObjectTablePanelFactory<T>( getIdStrategy(), configuration.getObjectTablePanelFactoryConfiguration( this ) );
  }

  private BeanAgent<T> agent;

  @NotNull
  public BeanAgent<T> getAgent() {
    if ( agent == null ) {
      agent = createAgent();
    }
    return agent;
  }

  @NotNull
  protected abstract BeanAgent<T> createAgent();

  @NotNull
  public LabelProvider getLabelProvider() {
    return configuration.getLabelProvider();
  }

  @NotNull
  public IdStrategy getIdStrategy() {
    return configuration.getIdStrategy();
  }

  @NotNull
  public Class<T> getType() {
    return configuration.getType();
  }

  @NotNull
  public AbstractConfiguration<T> getConfiguration() {
    return configuration;
  }

  @NotNull
  public BeanUiAccess<T> create() {
    return new DefaultBeanUiAccess<T>( getType(), getAgent(), getAccessAgent(), getBeansDisplay() );
  }

  /**
   * Abstract configuration
   *
   * @param <T> the configuration
   */
  public interface AbstractConfiguration<T> {
    /**
     * Returns the object table panel factory configuration
     *
     * @param builder
     * @return
     */
    @NotNull
    DefaultObjectTablePanelFactory.Configuration<T> getObjectTablePanelFactoryConfiguration( @NotNull UiAccessBuilder<T> builder );

    @NotNull
    Class<T> getType();

    @NotNull
    LabelProvider getLabelProvider();

    @NotNull
    IdStrategy getIdStrategy();

    @NotNull
    FormFactory<T> getFormFactory();
  }
}
