package com.cedarsoft.spring.rcp.binding;

import com.cedarsoft.utils.tags.Tag;
import com.cedarsoft.utils.tags.TagChangeListener;
import com.cedarsoft.utils.tags.TagObservable;
import com.cedarsoft.utils.tags.TagProvider;
import com.cedarsoft.utils.tags.TagSet;
import com.cedarsoft.utils.tags.ui.AbstractTagsComponentModel;
import com.cedarsoft.utils.tags.ui.TagsComponent;
import com.cedarsoft.utils.tags.ui.TagsComponentModel;
import org.jetbrains.annotations.NonNls;
import org.jetbrains.annotations.NotNull;
import org.springframework.binding.form.FormModel;
import org.springframework.binding.value.ValueModel;
import org.springframework.richclient.form.binding.Binding;
import org.springframework.richclient.form.binding.support.AbstractBinder;
import org.springframework.richclient.form.binding.support.AbstractBinding;
import org.springframework.richclient.list.ComboBoxAutoCompletion;
import org.springframework.richclient.list.EditableComboBoxAutoCompletion;

import javax.swing.JComboBox;
import javax.swing.JComponent;
import javax.swing.JLabel;
import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;
import java.util.List;
import java.util.Map;

/**
 * Binder for a tagset
 */
public class TagSetBinder extends AbstractBinder {
  @NotNull
  private final TagProvider tagProvider;

  /**
   * Creates a new binder for the given tag provider
   *
   * @param tagProvider the tag provider
   */
  public TagSetBinder( @NotNull TagProvider tagProvider ) {
    super( TagSet.class );
    this.tagProvider = tagProvider;
  }

  @Override
  protected JComponent createControl( Map context ) {
    return new JLabel( "dummy" );
  }

  @Override
  protected Binding doBind( @NotNull JComponent control, @NotNull FormModel formModel, @NotNull @NonNls String formPropertyPath, @NotNull Map context ) {
    return new TagSetBinding( tagProvider, formModel, formPropertyPath );
  }

  /**
   * Tag set binding
   */
  public static class TagSetBinding extends AbstractBinding {
    @NotNull @NonNls
    private static final String PROPERTY_VALUE = "value";

    @NotNull
    private final TagProvider tagProvider;

    /**
     * Creates a new binding
     *
     * @param tagProvider      the tag provider
     * @param formModel        the form model
     * @param formPropertyPath the property path
     */
    public TagSetBinding( @NotNull TagProvider tagProvider, @NotNull FormModel formModel, @NotNull @NonNls String formPropertyPath ) {
      super( formModel, formPropertyPath, TagSet.class );
      this.tagProvider = tagProvider;
    }

    @Override
    protected TagsComponent doBindControl() {
      final ValueModel tagsModel = getFormModel().getValueModel( formPropertyPath + ".tags", List.class );
      final TagSet copy = new TagSet( "copy" );

      final PropertyChangeListener changeListener = new PropertyChangeListener() {
        @Override
        public void propertyChange( PropertyChangeEvent evt ) {
          if ( PROPERTY_VALUE.equals( evt.getPropertyName() ) ) {
            copy.setTags( ( List<? extends Tag> ) tagsModel.getValue() );
          }
        }
      };
      tagsModel.addValueChangeListener( changeListener );

      copy.addTagChangeListener( new TagChangeListener() {
        @Override
        public void tagChanged( @NotNull TagChangeEvent event ) {
          tagsModel.setValueSilently( copy.getTags(), changeListener );
        }
      } );


      TagsComponentModel model = new AbstractTagsComponentModel( tagProvider ) {
        @Override
        public void selectTag( @NotNull Tag tag ) {
          copy.addTag( tag );
        }

        @Override
        public void unselectTag( @NotNull Tag tag ) {
          copy.removeTag( tag );
        }

        @NotNull
        @Override
        public TagObservable getSelectedTags() {
          return copy;
        }
      };

      TagsComponent component = new TagsComponent( model );
      {
        //Add the auto completion...
        JComboBox comboBox = component.getTagCombo();
        if ( comboBox.isEditable() ) {
          // It's editable, so install autocompletion for editable comboboxes
          //noinspection ResultOfObjectAllocationIgnored
          new EditableComboBoxAutoCompletion( comboBox );
        } else {
          //noinspection ResultOfObjectAllocationIgnored
          new ComboBoxAutoCompletion( comboBox );
        }
      }

      return component;
    }

    @Override
    protected void readOnlyChanged() {
    }

    @Override
    protected void enabledChanged() {
    }
  }
}
