package com.cedarsoft.spring.rcp.commands;

import com.cedarsoft.lookup.Lookup;
import com.cedarsoft.spring.rcp.selection.SelectionChangeListener;
import com.cedarsoft.spring.rcp.selection.SelectionModel;
import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;

/**
 * Support class that enables/disables a command based on a selection
 *
 * @param <T> the type of object that is selected
 */
public class SelectionAwareCommandSupport<T> {
  @NotNull
  private final Class<T> type;
  @NotNull
  private final SelectionModel selectionModel;
  @NotNull
  private final Condition<T> condition;

  /**
   * Creates a new selection aware command support
   *
   * @param type           the type of the selection
   * @param selectionModel the selection model
   * @param command        the command
   */
  protected SelectionAwareCommandSupport( @NotNull final Class<T> type, @NotNull SelectionModel selectionModel, @NotNull final ExtendedCommand command ) {
    this( type, selectionModel, command, null );
  }

  /**
   * Creates a new selection aware command support
   *
   * @param type           the type
   * @param selectionModel the selection model
   * @param command        the command
   * @param condition      the condition
   */
  protected SelectionAwareCommandSupport( @NotNull final Class<T> type, @NotNull SelectionModel selectionModel, @NotNull final ExtendedCommand command, @Nullable Condition<T> condition ) {
    this.type = type;
    this.selectionModel = selectionModel;
    if ( condition == null ) {
      this.condition = new Condition<T>() {
        @Override
        public boolean shallBeEnabled( @NotNull T selectedObject ) {
          return true;
        }
      };
    } else {
      this.condition = condition;
    }

    command.setEnabled( selectionModel.getCurrentlySelectedObject().lookup( type ) != null );
    selectionModel.addSelectionChangeListener( new SelectionChangeListener() {
      @Override
      public void selectionChanged( @NotNull Lookup selectedObjects ) {
        updateEnabledState( command );
      }
    } );
  }

  /**
   * Updates the enabled state
   *
   * @param command the command
   */
  protected void updateEnabledState( @NotNull ExtendedCommand command ) {
    T selectedObject = selectionModel.getCurrentlySelectedObject().lookup( type );
    if ( selectedObject != null ) {
      command.setEnabled( shallBeEnabled( selectedObject ) );
    } else {
      command.setEnabled( false );
    }
  }

  /**
   * May be overridden
   *
   * @param selectedObject the selected object
   * @return whether the command should be enabled
   */
  protected boolean shallBeEnabled( @NotNull T selectedObject ) {
    return condition.shallBeEnabled( selectedObject );
  }

  /**
   * Returns the currently selected object
   *
   * @return the currently selected object
   */
  @NotNull
  public T getSelectedObject() {
    T selectedObject = selectionModel.getCurrentlySelectedObject().lookup( type );
    if ( selectedObject == null ) {
      throw new IllegalStateException( "No object selected" );
    }
    return selectedObject;
  }

  /**
   * The condition
   */
  public interface Condition<T> {
    /**
     * Whether the condition shall be enabled
     *
     * @param selectedObject the selected object
     * @return whether it shall be enabled
     */
    boolean shallBeEnabled( @NotNull T selectedObject );

  }
}

