package com.cedarsoft.spring.rcp.creation;

import org.jetbrains.annotations.NonNls;
import org.jetbrains.annotations.NotNull;

import java.lang.reflect.Constructor;

/**
 * A creator model that uses reflection to create the object.
 * Call {@link #getFormObject()} after the form has been commited to create a new object.
 *
 * @param <T> the type this model creates.
 */
public class ReflectionCreatorModel<T> extends CreatorModel<T> {
  @NotNull
  private final Constructor<T> constructor;
  @NotNull
  @NonNls
  private final String[] argumentNames;

  /**
   * Creates a new model that will use a constructor with the same number of parameters as argumentNames has
   *
   * @param formId        the form id
   * @param type          the type
   * @param argumentNames the argument names
   */
  public ReflectionCreatorModel( @NotNull @NonNls String formId, @NotNull Class<T> type, @NonNls @NotNull String... argumentNames ) {
    this( formId, type, guessConstructor( type, argumentNames ), argumentNames );
  }

  /**
   * Creatse a new model
   *
   * @param formId        the form id
   * @param type          the type
   * @param constructor   the constructor that is used to instantiate the object
   * @param argumentNames the names of the arguments
   */
  public ReflectionCreatorModel( @NotNull @NonNls String formId, @NotNull Class<T> type, @NotNull Constructor<T> constructor, @NonNls @NotNull String... argumentNames ) {
    super( formId, type );
    this.constructor = constructor;
    this.argumentNames = argumentNames.clone();

    Class<?>[] types = constructor.getParameterTypes();
    if ( types.length != argumentNames.length ) {
      throw new IllegalStateException( "Had " + argumentNames.length + " argument names. Expected " + types.length );
    }

    for ( int i = 0; i < argumentNames.length; i++ ) {
      String argumentName = argumentNames[i];
      registerProperty( types[i], argumentName );
    }
  }

  @Override
  @NotNull
  protected T createObject() {
    Object[] arguments = new Object[argumentNames.length];

    for ( int i = 0; i < argumentNames.length; i++ ) {
      String argumentName = argumentNames[i];
      arguments[i] = getValueModel( argumentName ).getValue();
    }

    //noinspection OverlyBroadCatchBlock
    try {
      return constructor.newInstance( arguments );
    } catch ( Exception e ) {
      throw new RuntimeException( e );
    }
  }

  /**
   * Helper method that guesses a constructor for the given argument names.
   *
   * @param type          the type the constructor is searched for
   * @param argumentNames the argument names
   * @return the constructor if one is found.
   */
  @NotNull
  private static <T> Constructor<T> guessConstructor( @NotNull Class<T> type, @NotNull @NonNls String[] argumentNames ) {
    Constructor<T> found = null;

    Constructor<?>[] constructors = type.getConstructors();
    for ( Constructor<?> constructor : constructors ) {
      if ( constructor.getParameterTypes().length == argumentNames.length ) {
        if ( found != null ) {
          throw new IllegalStateException( "No unambiguous constructor found with " + argumentNames.length + " parameters for <" + type.getClass().getName() + '>' );
        }
        found = ( Constructor<T> ) constructor;
      }
    }

    if ( found == null ) {
      throw new IllegalStateException( "No constructor found for <" + type.getName() + "> with " + argumentNames.length + " parameters" );
    }

    return found;
  }
}
