package com.cedarsoft.spring.rcp.editor;

import com.cedarsoft.CanceledException;
import com.cedarsoft.ObjectCommit;
import com.cedarsoft.spring.rcp.renderer.DefaultLabelProvider;
import com.cedarsoft.spring.rcp.forms.DefaultFormUi;
import com.cedarsoft.spring.rcp.forms.ExtendedForm;
import com.cedarsoft.spring.rcp.forms.FormFactory;
import com.cedarsoft.spring.rcp.forms.FormPage;
import com.cedarsoft.spring.rcp.forms.FormUi;
import com.cedarsoft.spring.rcp.forms.StaticFormBackend;
import com.cedarsoft.spring.rcp.id.IdStrategy;
import org.jetbrains.annotations.NonNls;
import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;
import org.springframework.context.ApplicationEvent;
import org.springframework.richclient.selection.binding.support.LabelProvider;

import javax.swing.JComponent;
import javax.swing.JPanel;
import java.awt.BorderLayout;

/**
 * Editor based on a form
 */
public class FormBasedEditor<T> extends TitledEditor<T> {
  @NotNull
  @NonNls
  public static final String PROPERTY_OBJECT_COMMIT = "objectCommit";
  @NotNull
  @NonNls
  public static final String PROPERTY_FORM_FACTORY = "formFactory";
  @NotNull
  @NonNls
  public static final String PROPERTY_ID_STRATEGY = "idStrategy";
  @NotNull
  @NonNls
  public static final String PROPERTY_LABEL_PROVIDER = "labelProvider";
  @NotNull
  @NonNls
  public static final String PROPERTY_INITIALIZER_CALLBACK = "initializerCallback";

  @Override
  public String getDisplayName() {
    return getEditorObjectRepresentation();
  }

  @NotNull
  private String getEditorObjectRepresentation() {
    String objectLabel = getLabelProvider().getLabel( getEditorObject() );
    if ( objectLabel != null && objectLabel.length() > 0 ) {
      return objectLabel;
    }

    return getMessage( getIdStrategy().createEditorId() + ".title" );
  }

  @Override
  @NotNull
  protected String getTitle() {
    return getEditorObjectRepresentation();
  }

  @Override
  @NotNull
  protected String getMessage() {
    return getMessage( getIdStrategy().createEditorId() + ".message" );
  }

  @Override
  public boolean isDirty() {
    return getFormUiSupport().getForm().isDirty();
  }

  @Override
  public boolean canClose() {
    try {
      getFormUiSupport().commit( true );
    } catch ( CanceledException ignore ) {
      return false;
    }
    return true;
  }

  @Override
  protected void handleApplicationEvent( @NotNull ApplicationEvent event ) {
  }

  @Override
  public void initialize( Object editorObject ) {
    super.initialize( editorObject );

    InitializerCallback<T> callback = getInitializerCallback();
    if ( callback != null ) {
      callback.isInitializing( this );
    }
  }

  @NotNull
  @Override
  public String getId() {
    return getEditorObject().toString();
    //todo improve id...
  }

  private JPanel panel;

  @Override
  @NotNull
  protected JComponent createControl() {
    if ( panel == null ) {
      panel = new JPanel( new BorderLayout() );
      FormPage<T> formPage = new FormPage<T>( getFormUiSupport(), false );
      panel.add( formPage.getControl(), BorderLayout.CENTER );
    }
    return panel;
  }

  //The form ui support

  @Nullable
  private FormUi<T> formUi;

  @NotNull
  public FormUi<T> getFormUiSupport() {
    if ( formUi == null ) {
      StaticFormBackend<T> formBackend = new StaticFormBackend<T>( getIdStrategy().createFormId(), getEditorObject(), getObjectCommit() );
      formBackend.setShallAskForCommitConfirmation( true );
      formUi = new DefaultFormUi<T>( getIdStrategy().getBaseId(), formBackend, getFormFactory() );
    }

    return formUi;
  }

  //THE EDITOR PROPERTIES...

  @Nullable
  private IdStrategy idStrategy;
  @Nullable
  private FormFactory<T> formFactory;
  @Nullable
  private ObjectCommit<T> objectCommit;

  @Nullable
  private InitializerCallback<T> initializerCallback;

  @NotNull
  private LabelProvider labelProvider = new DefaultLabelProvider<T>() {
    @NotNull
    @Override
    protected String getLabelFor( @NotNull T item ) {
      return item.toString();
    }
  };

  @Nullable
  public InitializerCallback<T> getInitializerCallback() {
    return initializerCallback;
  }

  public void setInitializerCallback( @Nullable InitializerCallback<T> initializerCallback ) {
    this.initializerCallback = initializerCallback;
  }

  @NotNull
  public LabelProvider getLabelProvider() {
    return labelProvider;
  }

  public void setLabelProvider( @NotNull LabelProvider labelProvider ) {
    this.labelProvider = labelProvider;
  }

  @NotNull
  public ObjectCommit<T> getObjectCommit() {
    if ( objectCommit == null ) {
      throw new IllegalStateException( "No object commit set" );
    }
    return objectCommit;
  }

  public void setObjectCommit( @Nullable ObjectCommit<T> objectCommit ) {
    this.objectCommit = objectCommit;
  }

  @NotNull
  public IdStrategy getIdStrategy() {
    if ( idStrategy == null ) {
      throw new IllegalStateException( "No idStrategy set" );
    }
    return idStrategy;
  }

  @NotNull
  public FormFactory<T> getFormFactory() {
    if ( formFactory == null ) {
      throw new IllegalStateException( "formFactory not set" );
    }
    return formFactory;
  }

  public void setFormFactory( @Nullable FormFactory<T> formFactory ) {
    this.formFactory = formFactory;
  }

  public void setIdStrategy( @NotNull IdStrategy idStrategy ) {
    this.idStrategy = idStrategy;
  }

  @NotNull
  public ExtendedForm<? extends T> getForm() {
    return getFormUiSupport().getForm();
  }

  /**
   * An initializer callback that can be used to recieve lifecycle notifications.
   */
  public interface InitializerCallback<T> { //todo move to extended editor(?)

    void isInitializing( @NotNull FormBasedEditor<T> editor );
  }
}
