package com.cedarsoft.spring.rcp.forms;

import com.cedarsoft.CanceledException;
import org.jetbrains.annotations.NonNls;
import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;
import org.springframework.binding.form.FormModel;
import org.springframework.binding.form.ValidatingFormModel;

/**
 * Abstract base class for form backends that create the forms dynamically.
 *
 * @param <T> the type this form backend handles
 */
public abstract class AbstractFormBackend<T> implements FormBackend {
  private final boolean creatingNew;

  private boolean shallAskForCommitConfirmation;

  @NotNull
  private final FormModelFactory<T> formModelFactory;

  /**
   * Creates a new form backend
   *
   * @param formId the id for the form model (and form)
   * @param object the object. If set to null a new object is created.
   */
  protected AbstractFormBackend( @NotNull @NonNls String formId, @Nullable T object ) {
    this( formId, object, object == null );
  }

  /**
   * Creates a new form backend
   *
   * @param formId      the form id
   * @param object      the object
   * @param creatingNew whether a new object is created
   */
  protected AbstractFormBackend( @NotNull @NonNls String formId, @Nullable T object, boolean creatingNew ) {
    this( object, creatingNew, new DefaultFormModelFactory<T>( formId ) );
  }

  /**
   * Creates a new form backend.<br/>
   * Most of the time when the <code>object</code> is *not* null, <code>creatingNew</code> is set to false.
   *
   * @param object           the object that is edited within the form (may be null if a new object is created)
   * @param creatingNew      whether a new object is created (must be true if the object is null)
   * @param formModelFactory the form model factory
   */
  protected AbstractFormBackend( @Nullable T object, boolean creatingNew, @NotNull FormModelFactory<T> formModelFactory ) {
    this.object = object;
    this.creatingNew = creatingNew;
    this.formModelFactory = formModelFactory;
    if ( object == null && !creatingNew ) {
      throw new IllegalArgumentException( "Must provide an object if creatingNew is false" );
    }
  }

  /**
   * Do not use directly. Use {@link #getFormObject()} instead.
   */
  private T object;

  /**
   * Returns the object.
   * If {@link #isCreatingNew()} then the new object is created on the fly
   *
   * @return the object the form object
   *
   * @throws CanceledException if it has been canceled
   */
  @NotNull
  protected T getFormObject() throws CanceledException {
    if ( object == null ) {
      if ( isCreatingNew() ) {
        object = createNewObject();
      } else {
        throw new IllegalStateException( "No object available - but also not creating a new one????" );
      }
    }
    return object;
  }

  @Override
  public boolean isCreatingNew() {
    return creatingNew;
  }

  @Override
  public void commit() {
    if ( isCreatingNew() ) {
      addObject( getFormObject() );
    } else {
      commitObject( getFormObject() );
    }
  }

  /**
   * Returns the form model factory
   *
   * @return the form model factory
   */
  @NotNull
  public FormModelFactory<T> getFormModelFactory() {
    return formModelFactory;
  }

  /**
   * Creates the form model for the form object
   *
   * @return the created for model
   */
  @NotNull
  protected ValidatingFormModel createFormModel() {
    return formModelFactory.createFormModel( getFormObject() );
  }

  @Override
  public boolean shallAskForCommitConfirmation() {
    return shallAskForCommitConfirmation;
  }

  public void setShallAskForCommitConfirmation( boolean shallAskForCommitConfirmation ) {
    this.shallAskForCommitConfirmation = shallAskForCommitConfirmation;
  }

  /**
   * Do *not* access directly. Use {@link #getFormModel()} instead.
   */
  private FormModel formModel;

  /**
   * Creates a form model
   *
   * @return the form model
   */
  @Override
  @NotNull
  public FormModel getFormModel() {
    if ( formModel == null ) {
      formModel = createFormModel();
    }
    return formModel;
  }

  /**
   * Is called when the form has been commited and {@link #isCreatingNew()} returns true
   * Subclasses should add the object to the database (using a dao)
   *
   * @param addedObject the form object that has been created
   */
  protected abstract void addObject( @NotNull T addedObject );

  /**
   * Is called when the object has been edited
   *
   * @param commitedObject the object that has been edited
   */
  protected abstract void commitObject( @NotNull T commitedObject );

  /**
   * Creates the new object. The method {@link #addObject(Object)} is called
   * later when the form has been commited.
   *
   * @return the new object
   *
   * @throws CanceledException if it has been canceled
   */
  @NotNull
  protected abstract T createNewObject() throws CanceledException, UnsupportedOperationException;
}
