package com.cedarsoft.spring.rcp.selection;

import com.cedarsoft.spring.rcp.events.SelectionEvent;
import com.cedarsoft.lookup.DynamicLookup;
import com.cedarsoft.lookup.Lookup;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.jetbrains.annotations.NotNull;
import org.springframework.context.ApplicationEvent;
import org.springframework.context.ApplicationListener;

import java.util.ArrayList;
import java.util.List;
import java.util.Map;

/**
 * Contains the current selections
 */
public class SelectionModel implements ApplicationListener {
  private static final Log log = LogFactory.getLog( SelectionModel.class );

  @NotNull
  private final List<SelectionChangeListener> listeners = new ArrayList<SelectionChangeListener>();

  @NotNull
  private DynamicLookup currentlySelectedObjects = new DynamicLookup();

  @Override
  public void onApplicationEvent( @NotNull ApplicationEvent event ) {
    if ( event instanceof SelectionEvent ) {
      SelectionEvent selectionEvent = ( SelectionEvent ) event;

      if ( selectionEvent.getObjects().lookups().isEmpty() ) {
        return;
      }

      switch ( selectionEvent.getType() ) {
        case SELECT:
          select( selectionEvent.getObjects() );
          break;
        case UNSELECT:
          unselect( selectionEvent.getObjects() );
          break;
        default:
          throw new IllegalArgumentException( "Invalid type " + selectionEvent.getType() );
      }
    }
  }

  /**
   * Select the current object
   *
   * @param selectedObjects the selected objects
   */
  public void select( @NotNull Lookup selectedObjects ) {
    if ( log.isDebugEnabled() ) {
      log.debug( "-------- Selecting ----------" );
      for ( Map.Entry<Class<?>, ?> entry : selectedObjects.lookups().entrySet() ) {
        log.debug( "\t" + entry.getKey().getName() + ": " + entry.getValue() );
      }
      log.debug( "-------- /Selecting ----------" );
    }

    if ( currentlySelectedObjects.addValues( selectedObjects ) ) {
      notifySelectionChanged( this.currentlySelectedObjects );
    }
  }

  /**
   * Unselects the given objects
   *
   * @param unselectedObjects the objects that are unselected
   */
  public void unselect( @NotNull Lookup unselectedObjects ) {
    if ( log.isDebugEnabled() ) {
      log.debug( "-------- Unselecting ----------" );
      for ( Map.Entry<Class<?>, ?> entry : unselectedObjects.lookups().entrySet() ) {
        log.debug( "\t" + entry.getKey().getName() + ": " + entry.getValue() );
      }
      log.debug( "-------- /Unselecting ----------" );
    }

    if ( currentlySelectedObjects.removeValues( unselectedObjects ) ) {
      notifySelectionChanged( this.currentlySelectedObjects );
    }
  }

  public void selectionChanged( @NotNull Lookup unselects, @NotNull Lookup selects ) {
    currentlySelectedObjects.removeValues( unselects );
    currentlySelectedObjects.addValues( selects );
    notifySelectionChanged( currentlySelectedObjects );
  }

  private void notifySelectionChanged( @NotNull Lookup selectedObjects ) {
    if ( log.isDebugEnabled() ) {
      log.debug( "############### SELECTION CHANGED ######################" );
      log.debug( "Contains now:" );
      for ( Map.Entry<Class<?>, ?> entry : currentlySelectedObjects.lookups().entrySet() ) {
        log.debug( "\t" + entry.getKey().getName() + ": " + entry.getValue() );
      }
      log.debug( "#####################################" );
    }

    for ( SelectionChangeListener listener : listeners ) {
      listener.selectionChanged( selectedObjects );
    }
  }

  public void addSelectionChangeListener( @NotNull SelectionChangeListener listener ) {
    listeners.add( listener );
  }

  @NotNull
  public Lookup getCurrentlySelectedObject() {
    return currentlySelectedObjects;
  }
}
