package com.cedarsoft.spring.rcp.table;

import ca.odell.glazedlists.EventList;
import ca.odell.glazedlists.GlazedLists;
import ca.odell.glazedlists.SortedList;
import ca.odell.glazedlists.event.ListEvent;
import ca.odell.glazedlists.event.ListEventListener;
import ca.odell.glazedlists.gui.TableFormat;
import ca.odell.glazedlists.swing.EventSelectionModel;
import ca.odell.glazedlists.swing.TableComparatorChooser;
import ca.odell.glazedlists.util.concurrent.Lock;
import com.cedarsoft.ObjectAccess;
import com.cedarsoft.history.ElementsChangedEvent;
import com.cedarsoft.history.ElementsListener;
import com.cedarsoft.history.ObservableCollection;
import com.cedarsoft.spring.rcp.PropertyPath;
import com.cedarsoft.spring.rcp.table.renderer.DateTimeTableCellRenderer;
import com.cedarsoft.spring.rcp.table.renderer.EnumTableCellRenderer;
import com.cedarsoft.spring.rcp.table.renderer.IntervalTableCellRenderer;
import com.cedarsoft.spring.rcp.table.renderer.LocalDateTableCellRenderer;
import com.cedarsoft.spring.rcp.table.renderer.LocalTimeTableCellRenderer;
import com.cedarsoft.spring.rcp.table.renderer.PercentageTableCellRenderer;
import com.cedarsoft.spring.rcp.table.renderer.PeriodCellRenderer;
import com.cedarsoft.spring.rcp.table.renderer.TaggedCellRenderer;
import com.cedarsoft.spring.rcp.events.SelectionEvent;
import com.cedarsoft.spring.rcp.selection.NoElementSelectedException;
import com.cedarsoft.utils.tags.TagSet;
import com.cedarsoft.utils.tags.Tagged;
import org.jetbrains.annotations.NonNls;
import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;
import org.joda.time.DateTime;
import org.joda.time.Interval;
import org.joda.time.LocalDate;
import org.joda.time.LocalTime;
import org.joda.time.Period;
import org.springframework.beans.support.PropertyComparator;
import org.springframework.context.ApplicationEvent;
import org.springframework.context.ApplicationListener;
import org.springframework.richclient.application.event.LifecycleApplicationEvent;
import org.springframework.richclient.application.statusbar.StatusBar;
import org.springframework.richclient.command.ActionCommandExecutor;
import org.springframework.richclient.command.CommandGroup;
import org.springframework.richclient.core.Guarded;
import org.springframework.richclient.factory.AbstractControlFactory;
import org.springframework.richclient.table.TableUtils;
import org.springframework.richclient.table.support.GlazedTableModel;
import org.springframework.richclient.util.PopupMenuMouseListener;
import org.springframework.util.Assert;

import javax.swing.JComponent;
import javax.swing.JPopupMenu;
import javax.swing.JTable;
import javax.swing.ListSelectionModel;
import javax.swing.SwingUtilities;
import javax.swing.event.ListSelectionEvent;
import javax.swing.event.ListSelectionListener;
import javax.swing.table.TableColumnModel;
import java.awt.event.MouseAdapter;
import java.awt.event.MouseEvent;
import java.math.BigDecimal;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.Collections;
import java.util.Comparator;
import java.util.List;

/**
 * Extended tbject table that is based heavily on a {@link ObjectAccess}.
 *
 * @param <T> the type
 */
public class ExtendedObjectTable<T> extends ObjectTable<T> {
  @NotNull
  protected final ObjectAccess<? extends T> objectAccess;

  /**
   * Creates a new object table
   *
   * @param modelId           the model id
   * @param type              the type of objects that are shown within the table
   * @param objectAccess      the object access
   * @param columnDefinitions the column definitions
   */
  public ExtendedObjectTable( @NotNull @NonNls String modelId, @NotNull Class<? super T> type, @NotNull ObservableCollection<? extends T> objectAccess, @NotNull @NonNls ColumnDefinition... columnDefinitions ) {
    this( modelId, type, objectAccess, false, columnDefinitions );
  }

  /**
   * Creates a new object table
   *
   * @param modelId                 the model id
   * @param type                    the type
   * @param objectAccess            the object access
   * @param handleApplicationEvents whether application events should be handled by this tyble
   * @param columnDefinitions       the column definitions
   */
  public ExtendedObjectTable( @NotNull @NonNls String modelId, @NotNull Class<? super T> type, @NotNull ObjectAccess<? extends T> objectAccess, boolean handleApplicationEvents, @NotNull @NonNls ColumnDefinition... columnDefinitions ) {
    super( modelId, type, handleApplicationEvents, columnDefinitions );
    this.objectAccess = objectAccess;

    //Now add the listener if it is possible
    if ( !handleApplicationEvents && objectAccess instanceof ObservableCollection ) {
      ( ( ObservableCollection<T> ) objectAccess ).addElementListener( new ElementsListener<T>() {
        @Override
        public void elementsDeleted( @NotNull ElementsChangedEvent<? extends T> event ) {
          handleDeletedObjects( event.getElements() );
        }

        @Override
        public void elementsAdded( @NotNull ElementsChangedEvent<? extends T> event ) {
          List<? extends T> elements = event.getElements();
          handleNewObjects( elements );

          //now select the last element
          if ( !elements.isEmpty() ) {
            final T lastObject = elements.get( elements.size() - 1 );
            SwingUtilities.invokeLater( new Runnable() {
              @Override
              public void run() {
                selectObject( lastObject );
              }
            } );
          }
        }

        @Override
        public void elementsChanged( @NotNull ElementsChangedEvent<? extends T> event ) {
          handleUpdatedObjects( event.getElements() );
        }
      } );
    }
  }

  @NotNull
  @Override
  protected List<? extends T> getInitialData() {
    return objectAccess.getElements();
  }

  @NotNull
  public ObjectAccess<? extends T> getObjectAccess() {
    return objectAccess;
  }
}

