package com.cedarsoft.spring.rcp.tbpanel;

import com.cedarsoft.history.ObservableObjectAccess;
import com.cedarsoft.spring.rcp.id.IdStrategy;
import com.cedarsoft.spring.rcp.table.ObjectTable;
import com.cedarsoft.spring.rcp.tbpanel.aspects.ObjectTablePanelAspect;
import org.jetbrains.annotations.NonNls;
import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;

import java.util.List;

/**
 * Default table based panel factory.
 *
 * @param <T> the type for the panel
 */
public class DefaultObjectTablePanelFactory<T> implements ObjectTablePanelFactory<T> {
  @NotNull
  protected final IdStrategy idStrategy;
  @NotNull
  protected final Configuration<T> configuration;

  /**
   * Creates a new table based panel factory
   *
   * @param idStrategy    the id strategy
   * @param configuration the configuration
   */
  public DefaultObjectTablePanelFactory( @NotNull IdStrategy idStrategy, @NotNull Configuration<T> configuration ) {
    this.idStrategy = idStrategy;
    this.configuration = configuration;
  }

  @Override
  public ObjectTablePanel<T> createPanel( @NotNull ObservableObjectAccess<T> objectAccess ) {
    ObjectTablePanel<T> panel = new ObjectTablePanel<T>( idStrategy.createTablePanelId() );

    applyAspects( objectAccess, panel );

    panel.setObjectTable( configuration.createObjectTable( idStrategy.createTableId(), objectAccess ) );
    return panel;
  }

  protected void applyAspects( @NotNull ObservableObjectAccess<T> objectAccess, @NotNull ObjectTablePanel<T> panel ) {
    @Nullable
    List<? extends ObjectTablePanelAspect<T>> customAspects = configuration.getObjectTablePanelAspects( objectAccess, panel );
    if ( customAspects != null && !customAspects.isEmpty() ) {
      for ( ObjectTablePanelAspect<T> aspect : customAspects ) {
        panel.applyAspect( aspect );
      }
    }
  }

  /**
   * Configuration
   *
   * @param <T> the type
   */
  public interface Configuration<T> {
    /**
     * Creates an object table
     *
     * @param id           the id that shall be used for the table
     * @param objectAccess the object acces that shall be used
     * @return the created object table
     */
    @NotNull
    ObjectTable<T> createObjectTable( @NotNull @NonNls String id, @NotNull ObservableObjectAccess<T> objectAccess );

    /**
     * Returns the additional panel aspects (if there are any)
     *
     * @param objectAccess the object access
     * @param panel        the panel this aspects are created for
     * @return the panel configurer
     */
    @Nullable
    List<? extends ObjectTablePanelAspect<T>> getObjectTablePanelAspects( @NotNull ObservableObjectAccess<T> objectAccess, @NotNull ObjectTablePanel<T> panel );
  }
}
