package com.cedarsoft.spring.rcp.tbpanel.aspects;

import com.cedarsoft.ObjectAccess;
import com.cedarsoft.ObjectRemove;
import com.cedarsoft.PartTimeObjectAdd;
import com.cedarsoft.history.ObservableObjectAccess;
import com.cedarsoft.spring.rcp.aspects.AbstractAddRemoveAspect;
import com.cedarsoft.spring.rcp.beans.BeanAgent;
import com.cedarsoft.spring.rcp.commands.CommandIds;
import com.cedarsoft.spring.rcp.commands.CommandProvider;
import com.cedarsoft.spring.rcp.commands.DeleteCommand;
import com.cedarsoft.spring.rcp.commands.PartTimeObjectAccessCommandConfigurer;
import com.cedarsoft.spring.rcp.id.DefaultIdStrategy;
import com.cedarsoft.spring.rcp.id.IdStrategy;
import com.cedarsoft.spring.rcp.table.ObjectTable;
import com.cedarsoft.spring.rcp.tbpanel.ObjectTablePanel;
import com.cedarsoft.spring.rcp.tbpanel.SidePanelCommandsProvider;
import com.cedarsoft.spring.rcp.tbpanel.config.PopupConfigurer;
import org.jetbrains.annotations.NonNls;
import org.jetbrains.annotations.NotNull;
import org.springframework.binding.value.ValueModel;
import org.springframework.richclient.application.PageComponentContext;
import org.springframework.richclient.command.ActionCommand;
import org.springframework.richclient.command.support.GlobalCommandIds;
import org.springframework.richclient.list.ListMultipleSelectionGuard;

import javax.swing.JPanel;
import java.util.ArrayList;
import java.util.Collection;
import java.util.List;

/**
 * @param <T> the type of the bean this aspect is able to add/remove
 */
public class AddRemoveAspect<T> extends AbstractAddRemoveAspect<ObjectTablePanel<T>> implements ObjectTablePanelAspect<T> {
  /**
   * Factory method
   *
   * @param idStrategy   the id strategy
   * @param agent        the agent
   * @param objectAccess the object access
   * @param parent       the parent
   * @return the created add/remove aspect
   */
  @NotNull
  public static <T> AddRemoveAspect<T> create( @NotNull IdStrategy idStrategy, @NotNull BeanAgent<T> agent, @NotNull ObservableObjectAccess<T> objectAccess, @NotNull JPanel parent ) {
    //Add/Remove
    AddRemoveAspect<T> addRemoveAspect = new AddRemoveAspect<T>( idStrategy, new DefaultAddRemoveCallback<T>( objectAccess, agent, parent ) );

    //if the object access is instance of PartTimeObjectAdd
    if ( objectAccess instanceof PartTimeObjectAdd ) {
      addRemoveAspect.getNewCommandBuilder().setConfigurer( new PartTimeObjectAccessCommandConfigurer( ( PartTimeObjectAdd<?> ) objectAccess ) );
    }
    return addRemoveAspect;
  }

  /**
   * Creates a new aspect
   *
   * @param baseId   the base id
   * @param callback the callback
   */
  public AddRemoveAspect( @NotNull @NonNls String baseId, @NotNull Callback<T> callback ) {
    this( new DefaultIdStrategy( baseId ), callback );
  }

  /**
   * Creates a new add/remove aspect
   *
   * @param idStrategy the id strategy
   * @param callback   the callback
   */
  public AddRemoveAspect( @NotNull final IdStrategy idStrategy, @NotNull final Callback<T> callback ) {
    super( idStrategy );

    //new command
    getNewCommandBuilder().setProvider( new CommandProvider<ObjectTablePanel<T>>() {
      @Override
      @NotNull
      public ActionCommand getCommand( @NotNull ObjectTablePanel<T> object ) {
        return new ActionCommand( NEW_COMMAND_ID ) {
          @Override
          protected void doExecuteCommand() {
            callback.createNewEntry();
          }
        };
      }
    } );

    //Delete command
    getDeleteCommandBuilder().setProvider( new CommandProvider<ObjectTablePanel<T>>() {
      @Override
      @NotNull
      public ActionCommand getCommand( @NotNull final ObjectTablePanel<T> object ) {
        return new DeleteCommand<T>( new ObjectAccess<T>() {
          @Override
          @NotNull
          public List<? extends T> getElements() {
            return object.getObjectTable().getSelectedObjects();
          }
        }, new ObjectRemove<T>() {
          @Override
          public void remove( @NotNull T element ) {
            callback.removeEntry( element );
          }
        }, getIdStrategy() );
      }
    } );
  }

  @Override
  public void apply( @NotNull ObjectTablePanel<T> panel ) {
    final ActionCommand theNewCommand = getNewCommand( panel );
    final ActionCommand theDeleteCommand = getDeleteCommand( panel );

    panel.addConfigurer( new PopupConfigurer<T>() {
      @Override
      @NotNull
      protected List<? extends ActionCommand> getCommands( @NotNull ObjectTable<T> objectTable ) {
        List<ActionCommand> commands = new ArrayList<ActionCommand>();

        if ( theNewCommand != null ) {
          commands.add( theNewCommand );
        }

        if ( theDeleteCommand != null ) {
          commands.add( theDeleteCommand );
        }
        return commands;
      }
    } );

    panel.addCommandProvider( new SidePanelCommandsProvider() {
      @Override
      @NotNull
      public Collection<? extends ActionCommand> getCommands( @NotNull ValueModel selectionHolder ) {
        List<ActionCommand> commands = new ArrayList<ActionCommand>();

        if ( theNewCommand != null ) {
          commands.add( theNewCommand );
        }

        if ( theDeleteCommand != null ) {
          commands.add( theDeleteCommand );

          //noinspection ResultOfObjectAllocationIgnored
          new ListMultipleSelectionGuard( selectionHolder, theDeleteCommand );
        }

        return commands;
      }

      @Override
      public void registerLocalCommandExecutors( @NotNull PageComponentContext context ) {
        if ( theNewCommand != null ) {
          context.register( CommandIds.NEW, theNewCommand );
        }

        if ( theDeleteCommand != null ) {
          context.register( GlobalCommandIds.DELETE, theDeleteCommand );
        }
      }
    } );
  }

  /**
   * The callback for this aspect
   *
   * @param <T> the type of the entry that is added/removed
   */
  public interface Callback<T> {
    /**
     * Creates a new entry. This method must publish an event describing the addition of the newly
     * created entry.
     */
    void createNewEntry();

    /**
     * Is called when the user tried to delete an entry.
     * This method is called within a background thread
     *
     * @param entry the entry that is deleted
     */
    void removeEntry( @NotNull T entry );
  }
}
