package com.cedarsoft.spring.rcp.tree.layered;

import com.cedarsoft.spring.rcp.tree.ExtendedTreeModel;
import com.cedarsoft.spring.rcp.tree.ObservableCollectionBridge;
import com.cedarsoft.spring.rcp.tree.TreeUtils;
import com.cedarsoft.NotFoundException;
import com.cedarsoft.history.ElementsListener;
import com.cedarsoft.history.ObservableCollection;
import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;

import javax.swing.tree.DefaultMutableTreeNode;

/**
 * A tree model based on a layers configuration
 *
 * @param <R> the root objects type
 */
public class LayeredTreeModel<R> extends ExtendedTreeModel {
  @NotNull
  private final LayersConfiguration layersConfiguration;

  public LayeredTreeModel( @NotNull R root, @NotNull final LayersConfiguration layersConfiguration ) {
    super( new DefaultMutableTreeNode( root ) );
    this.layersConfiguration = layersConfiguration;

    TreeLayer<Object, Object> treeLayer = ( TreeLayer<Object, Object> ) layersConfiguration.getLayer( root.getClass() );
    ObservableCollection<Object> children = treeLayer.getChildren( root );

    MyCallback callback = new MyCallback();

    ObservableCollectionBridge<Object> rootBridge = new ObservableCollectionBridge<Object>( this, getRoot(), callback );
    rootBridge.install( children );
  }

  @Override
  public boolean isLeaf( @NotNull Object node ) {
    Object userObject = TreeUtils.getUserObject( node );
    return getLayerConfiguration( userObject ).isLeaf( userObject );
  }

  /**
   * Returns the layer configuration for the given parent
   *
   * @param parent the parent
   * @param <P>    the parent type
   * @param <C>    the type of the children
   * @return the layer configuration
   * @throws NotFoundException
   */
  @NotNull
  protected <P, C> TreeLayer<P, C> getLayerConfiguration( @NotNull P parent ) throws NotFoundException {
    return ( TreeLayer<P, C> ) layersConfiguration.getLayer( parent.getClass() );
  }

  private class MyCallback implements ObservableCollectionBridge.Callback<Object> {
    @Override
    @Nullable
    public ElementsListener<?> handleAddition( @NotNull Object element, @NotNull DefaultMutableTreeNode node ) {
      TreeLayer<Object, Object> config = getLayerConfiguration( element );

      ObservableCollection<Object> children = config.getChildren( element );
      if ( children == null ) {
        return null;
      }

      //Recursion
      ObservableCollectionBridge<Object> childBridge = new ObservableCollectionBridge<Object>( LayeredTreeModel.this, node, this );
      childBridge.install( children );

      return childBridge;
    }

    @Override
    public void handleDeleted( @NotNull Object element, @Nullable ElementsListener<?> storedListener ) {
      if ( storedListener == null ) {
        return;
      }

      TreeLayer<Object, Object> treeLayer = getLayerConfiguration( element );

      ObservableCollection<Object> children = treeLayer.getChildren( element );
      if ( children == null ) {
        throw new IllegalStateException( "No children found for deregistration at " + element );
      }
      children.removeElementListener( ( ElementsListener<? super Object> ) storedListener );
    }
  }
}
